<?php

declare(strict_types=1);

namespace hood\vk\api;

use hood\vk\robots\robot;

use Throwable;
use Exception;
use ArrayAccess;

/**
 * Настройки соединения с API
 *
 * @var robot $robot Робот
 * @var array $settings Настройки
 * @var float $version Версия API
 *
 * @todo
 * 1. Создать __isset(), __unset()
 */
class settings implements ArrayAccess
{
    /**
     * Конструктор
     */
    public function __construct(
        protected robot $robot,
        protected array $settings = []
    ) {
        if (empty($settings)) {
            // Настройки не получены

            // Инициализация
            $this->_init();
        }
    }

    /**
     * Инициализация (безопасная)
     *
     * @var float|null $version Версия API
     */
    public function init(float|null $version = null): self
    {
        if (isset($blocked)) {
            // Блокировка найдена

            throw new Exception('Повторная инициализация запрещена', 500);
        }

        // Блокировка
        static $blocked = true;

        // Инициализация
        try {
            $this->_init($version);
        } catch (Throwable $t) {
            throw new Exception('Не удалось инициализировать API', 500, $t->getPrevious());
        }

        return $this;
    }

    /**
     * Реинициализация
     *
     * @var float|null $version Версия API
     */
    public function reinit(float|null $version = null): self
    {
        // Буфер
        $version = $version ?? $this->settings['v'] ?? null;

        // Деинициализация
        $this->settings = [];

        // Инициализация
        try {
            $this->_init($version);
        } catch (Throwable $t) {
            throw new Exception('Не удалось инициализировать API', 500, $t->getPrevious());
        }

        return $this;
    }

    /**
     * Инициализация
     *
     * @var float $version Версия API
     */
    protected function _init(float $version = 5.124): self
    {
        // Ключ
        $this->settings['access_token'] = $this->robot->key;

        // Версия API
        $this->settings['v'] = $version;

        return $this;
    }

    /**
     * Записать получателя
     *
     * Определяет получателей по входным параметрам
     *
     * @see hood\vk\api\methods\messages Сообщения
     */
    public function writeDestination(string|array|int $destination): void
    {
        if (is_int($destination)) {
            // Идентификатор

            $this->settings['peer_id'] = $destination;
        } else if (is_array($destination)) {
            // Идентификаторы

            $this->settings['user_ids'] = $destination;
        } else if (is_string($destination)) {
            // Домен

            $this->settings['domain'] = $destination;
        }

        throw new Exception('Не удалось определить получателя', 500);
    }

    /**
     * Записать свойство
     *
     * @param string $name Название
     * @param mixed $value Значение
     */
    public function __set(string $name, mixed $value): void
    {
        match ($name) {
            'settings' => isset($this->settings) ? throw new Exception('Запрещено перезаписывать настройки', 500) : $this->settings = $value,
            'robot' => isset($this->robot) ? throw new Exception('Запрещено перезаписывать Робота', 500) : $this->robot = $value,
            'data' => $this->offsetSet('data', $value),
            'attachments' => $this->offsetSet('attachments', $value),
            default => $this->offsetSet($name, $value)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param string $name Название
     */
    public function __get(string $name): mixed
    {
        return match ($name) {
            'settings' => $this->settings ?? throw new Exception('Настройки не инициализированы', 500),
            'robot' => $this->robot ?? throw new Exception('Робот не инициализирован', 500),
            'data' => $this->offsetGet('data'),
            'attachments' => $this->offsetGet('attachments'),
            default => $this->offsetGet($name)
        };
    }

    /**
     * Записать по смещению
     */
    public function offsetSet(mixed $offset, mixed $value): mixed
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полная запись

                throw new Exception('Запрещено перезаписывать настройки', 500);
            } else if (
                (strcasecmp($offset, 'data') === 0)
                || (strcasecmp($offset, 'attachment') === 0)
                || (strcasecmp($offset, 'attachments') === 0)
            ) {
                // Записать вложения

                if (is_array($value)) {
                    // Жесткая запись

                    return $this->settings['attachments'] = $value;
                }

                // Конкатенация
                return $this->settings['attachments'][] = $value;
            } else {
                // Запись по ключу или смещению

                return $this->settings[$offset] = $value;
            }
        }

        throw new Exception('Настройки не инициализированы', 500);
    }

    /**
     * Прочитать по смещению
     */
    public function &offsetGet(mixed $offset): mixed
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полное чтение

                return $this->settings;
            } else if (strcasecmp($offset, 'data') === 0 || strcasecmp($offset, 'attachments') === 0) {
                // Прочитать вложения

                return $this->settings['attachments'];
            } else if (array_key_exists($offset, $this->settings)) {
                // Прочитать по ключу или смещению

                return $this->settings[$offset];
            }

            is_int($offset) ? throw new Exception("Смещение $offset не найдено", 404) : throw new Exception("Ключ $offset не найден", 404);
        }

        throw new Exception('Настройки не инициализированы', 500);
    }

    /**
     * Проверка существования смещения
     */
    public function offsetExists(mixed $offset): bool
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полная проверка

                return isset($this->settings);
            } else if (strcasecmp($offset, 'data') === 0 || strcasecmp($offset, 'attachments') === 0) {
                // Проверка вложений

                return array_key_exists('attachments', $this->settings);
            } else {
                // Проверка по ключу или смещению

                return array_key_exists($offset, $this->settings);
            }
        }

        throw new Exception('Настройки не инициализированы', 500);
    }

    /**
     * Удалить по смещению
     */
    public function offsetUnset(mixed $offset): void
    {
        if (isset($this->settings)) {
            if (strcasecmp($offset, 'settings') === 0) {
                // Полное удаление

                unset($this->settings);

                return;
            } else if (strcasecmp($offset, 'data') === 0 || strcasecmp($offset, 'attachments') === 0) {
                // Удаление вложений

                unset($this->settings['attachments']);

                return;
            } else {
                // Удаление по ключу или смещению

                unset($this->settings[$offset]);

                return;
            }
        }

        throw new Exception('Настройки не инициализированы', 500);
    }
}
